# -*- coding: utf-8 -*-

"""
# Colonia de hormigas
```
Versión   : 1.2, Para uso educativo
Autor     : Luis Beltran Palma Ttito
Lugar     : Cusco, Perú, 2024.
Proposito : Implementación de colonia de hormigas en python
Problema  : TSP.
```

#Librerías
"""

import random
import math
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

"""# Clase Algorítmo Genético"""

    
class ColoniaHormiga:
    # Constructor
    def __init__(self, Iteraciones, CantHormigas, Ro, Q):
        self.Iteraciones = Iteraciones          # Cantidad de iteraciones
        self.CantHormigas = CantHormigas        # Cantidad de hormigas
        self.Ro = Ro                            # Coeficiente de evaporación
        self.Q = Q                              # Coeficiente de disipación
        self.Archivo = ''                       # Archivo TSP
        self.TSP = []                           # TSP
        self.Distancia = []                     # Matriz de distancias
        self.Visibilidad = []                   # matriz de visibilidad de aristas por parte de las hormigas
        self.Feromona = []                      # matriz de feromona
        self.N = 0                              # Tamaño de la matriz, N x N

        self.Mejor = []                         # Aqui almacena la mejor solución en cada iteracion
        self.HistCosto = []                     # Historico de solucion
        self.Ruta = []                          # Ruta solucion
        self.Costo = 0.0                        # Costo de la mejor solución

    # Muestra evolución de costo
    def GraficaCosto(self):
        plt.figure()
        plt.plot(self.HistCosto)
        plt.title('Evolución de costo')
        plt.grid(True)
        plt.show

        # Grafica la ruta solución 2D
    def GraficaRuta(self):
        plt.figure()
        X = []
        Y = []
        for i in range(len(self.Ruta)):
            X.append(self.TSP[self.Ruta[i]-1, 0])
            Y.append(self.TSP[self.Ruta[i]-1, 1])
        plt.plot(X,Y,'-o')
        for i in range(len(self.Ruta)):
            plt.annotate(str(self.Ruta[i]), (X[i],Y[i]))
        plt.title("Ruta solución")
        plt.show()

    # Lee Matriz desde excel (sin cabecera) y crea matriz de distancias
    def LeerDistanciaExcel(self, Archivo, Hoja ):
        df = pd.read_excel(Archivo, sheet_name=Hoja, header=None)
        df_float = df.astype(float)
        self.Distancia = np.array(df_float.values)
        # Reemplazar valores iguales a cero por 0.000001, para evitar división por cero
        self.Distancia[self.Distancia == 0] = 0.000001
        self.N = len(self.Distancia)
        self.Mejor = list(range(1,self.N+1))
        self.Visibilidad = 1/self.Distancia
        self.Feromona = self.Distancia * 0.0 + 0.1
        return self.N, self.TSP, self.Distancia, self.Visibilidad, self.Feromona
    
    # Lee archivo TSP descargado de TSPLIB
    def Leer_TSP(self, Archivo):
        with open(Archivo) as archivo:
            ciudades = []
            while True:
                linea = archivo.readline()
                if linea.startswith("NODE_COORD_SECTION"):
                    break
            for linea in archivo:
                partes = linea.strip().split()
                if partes[0] == "EOF":
                    break
                _, x, y = partes
                ciudades.append([float(x), float(y)])
        return np.array(ciudades)

    # Lee Archivo TSP y crea matriz de distancias
    def LeerTSP(self, Archivo):
        # Leer TSP
        self.TSP = self.Leer_TSP(Archivo)
        self.N = len(self.TSP)
        self.Mejor = list(range(1,self.N+1))
        # Convertir TSP a matriz de distancias
        self.Distancia = np.empty((self.N,self.N))
        for i in range(self.N):
            for j in range(self.N):
                if i == j:
                    self.Distancia[i, j] = 0.000001 # debiera ser 0, pero 1/0 genera división por cero, pero no sera utilizado 
                else:
                    self.Distancia[i, j] = math.sqrt(math.pow(self.TSP[i, 0] - self.TSP[j, 0], 2)
                    + math.pow(self.TSP[i, 1] - self.TSP[j, 1], 2))
        # Cálculo de visibilidad y feromona
        self.Visibilidad = 1/self.Distancia
        self.Feromona = self.Distancia * 0.0 + 0.1
        return self.N, self.TSP, self.Distancia, self.Visibilidad, self.Feromona

    # Calcula costo entre 2 nodos adyacentes
    def CostoArco(self, inicio, fin):
        return self.Distancia[inicio-1][fin-1]

    # Calcula el costo del camino C
    def CostoRuta(self,C):
        Cost = 0
        for k in range(self.N-1):
            Cost = Cost + self.CostoArco(C[k],C[k+1])
        Cost = Cost + self.CostoArco(C[self.N-1], 1)
        return Cost

    # Una hormina busca un camino a la comida aleatoriamente usando ruleta
    # Genera ruleta para cada adyacente desde un Nodo en  base a visibilidad, feromona
    # Solo crear caminos desde el Nodo actual a los faltantes (ADAPTACIÓN VIAJERO COMERCIO)
    def Hormiga(self):
        R = []  # R almacena la ruta que determinara la hormiga
        S = list(range(2,self.N+1))  # siempre inicia en el nodo 1,--> elegira los otros nodos al azar, usando ruleta por feromona
        Nodo = 1
        R.append(Nodo)

        # los elementos de S van eliminandose al azar, hasta que quede vacio
        while (len(S) > 0):
            L = [] # L contiene la matriz de la ruleta

            # Calcula el sumatoria de visibilidad * feromona
            Total = 0.0
            for k in S:
                Total = Total + self.Visibilidad[Nodo-1][k-1] * self.Feromona[Nodo-1][k-1]

            # calcula la función de probabilidad y función de distribución (genera ruleta)
            FD = 0.0
            for k in S:
                FD = FD + self.Visibilidad[Nodo-1][k-1] * self.Feromona[Nodo-1][k-1]/Total
                Fila = [Nodo, k, self.Visibilidad[Nodo-1][k-1] * self.Feromona[Nodo-1][k-1],
                        self.Visibilidad[Nodo-1][k-1] * self.Feromona[Nodo-1][k-1]/Total, FD]
                L.append(Fila)

            # Genera un valor aleatorio r y determina el siguiente nodo a elegir
            k = 0
            r = random.random()
            while (L[k][4] < r):
                k = k + 1
            Nodo = L[k][1]
            S.remove(Nodo)  # elimina nodo a elegir de las lista de nodos pendientes
            R.append(Nodo)  # almacena el nodo siguiente en la ruta

        # calcula costo de la ruta R
        C = self.CostoRuta(R)
        return R, C

    # N hormigas buscan ruta hacia la comida
    # y se va registrando la mejor solución
    def BuscanComida(self, N):
        Soluciones = []
        for k in range(N):
            Sol, _ = self.Hormiga()
            if self.CostoRuta(Sol) < self.CostoRuta(self.Mejor):
                self.Mejor = Sol
            Soluciones.append(Sol)
        return Soluciones, self.Mejor

    # Verifica si hay camino de p a q en el camino C
    # util para recalcular feromonas
    def HayCamino(self, C, p, q):
        for i in range(self.N - 1):
            if ((C[i] == p) and (C[i + 1] == q)):
                return True
        return False

    # Recalcula valor de feromona despues que N hormigas generaron caminos
    # nuevo valor feromona = f(valor evaporacion de feromona, suma deltas de cada hormiga)
    def RecalculaFeromona(self, Soluciones,Ro,Q):

        # disipación feromona
        for i in range(0, self.N):
            for j in range(0, self.N):
                self.Feromona[i][j] = (1 - Ro) * self.Feromona[i][j]

        # incremento feromona
        for i in range(0, self.N - 1):
            for j in range(i + 1, self.N):
                SumDelta = 0
                # print('(', i + 1, ',', j + 1, ')', end=' ')
                for R in Soluciones:
                    Cost = self.CostoRuta(R)
                    if (self.HayCamino(R, i + 1, j + 1)):
                        SumDelta = SumDelta + Q/Cost
                self.Feromona[i][j] = self.Feromona[i][j] + SumDelta
        return self.Feromona

    # Ejecuta el algoritmo de colonia de hormigas
    # Iteraciones : cantidad de iteraciones a realizar, es decir cantidad de veces que las hormigas salen en busca de alimentos
    # CantHormigas: cantidad de hormigas que salen a buscar alimentos en cada iteración
    # Ro: porcentaje de evaporación de feromoma
    # Q: Coef. de aprendizaje
    def EjecutarTSP(self, Archivo):
        self.Archivo = Archivo
        self.LeerTSP(self.Archivo)
        self.HistCosto = []
        for k in range(self.Iteraciones):
            Soluciones, self.Mejor = self.BuscanComida(self.CantHormigas)
            self.RecalculaFeromona(Soluciones,self.Ro,self.Q)
            self.HistCosto.append(self.CostoRuta(self.Mejor))
        self.Ruta = self.Mejor
        self.Costo = self.CostoRuta(self.Mejor)
        return self.Ruta, self.Costo

    def EjecutarDistancia(self, Archivo, Hoja):
        self.Archivo = Archivo
        self.LeerDistanciaExcel(self.Archivo, Hoja)
        self.HistCosto = []
        for k in range(self.Iteraciones):
            Soluciones, self.Mejor = self.BuscanComida(self.CantHormigas)
            self.RecalculaFeromona(Soluciones,self.Ro,self.Q)
            self.HistCosto.append(self.CostoRuta(self.Mejor))
        self.Ruta = self.Mejor
        self.Costo = self.CostoRuta(self.Mejor)
        return self.Ruta, self.Costo



# -*- coding: utf-8 -*-

"""
#Algoritmo Genético
```
Versión   : 1.2, Para uso educativo
Autor     : Luis Beltran Palma Ttito
Lugar     : Cusco, Perú, 2024.
Proposito : Implementación de algoritmos genético en python
Problema  : TSP.
```

#Librerías
"""

import random
import math
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

"""# Clase Algorítmo Genético"""

    
class AlgoritmoGenetico:
    # Constructor
    def __init__(self, Generaciones, Poblacion, ProbMutacion, PobElite, funFitness=None):
        self.N = 0                              # Cantidad de ciudades
        self.Generaciones = Generaciones        # Cantidad de generaciones
        self.P = Poblacion                      # población
        self.ProbMutacion = ProbMutacion        # Probabilidad de mutación
        self.PoElite = PobElite                 # Cantidad de individuos de la élite
        self.Archivo = ''                       # Archivo TSP

        self.HistCosto = []         # Historial de costo
        self.HistFitness = []       # historial de fitness
        self.TSP = []               # TSP
        self.Ruta = []              # Ruta solución
        self.Distancia = []         # Matriz de distancias
        self.Costo = 0.0            # Costo de la ruta solucion
        self.Aptitu = 0.0           # Fitness de ruta solución
        self.Fitness = funFitness or self.Fitness_defecto # Función fitness definido por usuario

        
    # Muestra evolución de costo
    def GraficaCosto(self):
        plt.figure()
        plt.plot(self.HistCosto)
        plt.title('Evolución de costo')
        plt.grid(True)
        plt.show
      
    # Muestra evolución de aptitud
    def GraficaAptitud(self):
        plt.figure()
        plt.plot(self.HistFitness)
        plt.title('Evolución de aptitud')
        plt.grid(True)
        plt.show

    # Grafica la ruta solución 2D
    def GraficaRuta(self):
        plt.figure()
        X = []
        Y = []
        for i in range(len(self.Ruta)):
            X.append(self.TSP[self.Ruta[i]-1, 0])
            Y.append(self.TSP[self.Ruta[i]-1, 1])
        plt.plot(X,Y,'-o')
        for i in range(len(self.Ruta)):
            plt.annotate(str(self.Ruta[i]), (X[i],Y[i]))
        plt.title("Ruta solución")
        plt.show()

    # Lee Matriz desde excel (sin cabecera) y crea matriz de distancias
    def LeerDistanciaExcel(self, Archivo, Hoja ):
        # Leer TSP
        df = pd.read_excel(Archivo, sheet_name=Hoja, header=None)
        df_float = df.astype(float)
        self.Distancia = np.array(df_float.values)
        self.N = len(self.Distancia)
        return self.N, self.TSP, self.Distancia
    
    # Lee archivo TSP descargado de TSPLIB
    def Leer_TSP(self, Archivo):
        with open(Archivo) as archivo:
            ciudades = []
            while True:
                linea = archivo.readline()
                if linea.startswith("NODE_COORD_SECTION"):
                    break
            for linea in archivo:
                partes = linea.strip().split()
                if partes[0] == "EOF":
                    break
                _, x, y = partes
                ciudades.append([float(x), float(y)])
        return np.array(ciudades)
  
      
    # Lee Archivo TSP y crea matriz de distancias
    def LeerTSP(self, Archivo):
        # Leer TSP
        self.TSP = self.Leer_TSP(Archivo)
        self.N = len(self.TSP)
        # Convertir TSP a matriz de distancias
        self.Distancia = np.empty((self.N,self.N))
        for i in range(self.N):
            for j in range(self.N):
                if i == j:
                    self.Distancia[i, j] = 0
                else:
                    self.Distancia[i, j] = math.sqrt(math.pow(self.TSP[i, 0] - self.TSP[j, 0], 2)
                    + math.pow(self.TSP[i, 1] - self.TSP[j, 1], 2))

        return self.N, self.TSP, self.Distancia
      
    # Lee TSP excel (sin cabecera) y crea matriz de distancias
    def LeerExcel(self, Archivo, Hoja ):
        # Leer TSP
        df = pd.read_excel(Archivo, sheet_name=Hoja, header=None)
        self.TSP = np.array(df.values)
        self.N = len(self.TSP)
        # Convertir TSP a matriz de distancias
        self.Distancia = np.empty((self.N,self.N))
        for i in range(self.N):
            for j in range(self.N):
                if i == j:
                    self.Distancia[i, j] = 0
                else:
                    self.Distancia[i, j] = math.sqrt(math.pow(self.TSP[i, 0] - self.TSP[j, 0], 2)
                    + math.pow(self.TSP[i, 1] - self.TSP[j, 1], 2))

        return self.N, self.TSP, self.Distancia

    # Calcula Costo entre 2 nodos adyacentes
    def CostoArco(self, inicio, fin):
        return self.Distancia[inicio - 1][fin - 1]

    # Calcula Costo de una ruta
    def DistaTotal(self, R):
        Costo = 0
        for k in range(0, self.N - 1):
            Costo = Costo + self.CostoArco(R[k], R[k+1])
        Costo = Costo + self.CostoArco(R[self.N - 1], 1)
        return Costo

    # Calcula aptitup o fitnnes de la ruta R = 1/sum(distancias)
    def Fitness_defecto(self, R):
        Costo = 0
        for k in range(0, self.N - 1):
            Costo = Costo + self.CostoArco(R[k], R[k+1])
        Costo = Costo + self.CostoArco(R[self.N - 1], 1)
        return 1/Costo

    # Genera población inicial con P individuos
    def GeneraPoblacionInicial(self, P):
        self.Poblacion = []
        self.P = P
        for i in range(1, P + 1):
            S = random.sample(list(range(2,self.N+1)), k=self.N-1)
            S.insert(0, 1)
            self.Poblacion.append([self.Fitnnes(S),S])
        return self.Poblacion

    # Seleccion de padre y madre por metodo de la ruleta
    def Seleccion(self):

        # ordenar población por el valor fitness
        self.Poblacion = sorted(self.Poblacion)

        # Hallar sumatoria de fitness para crear ruleta
        SumaF = 0
        for P in self.Poblacion:
            SumaF = SumaF + P[0]

        # Crea ruleta
        Ruleta = []
        SumaAcu = 0
        for P in self.Poblacion:
            SumaAcu = SumaAcu + P[0]/SumaF
            Ruleta.append([P[0], P[1], SumaAcu])

        # Seleccionar padre
        RandPadre = random.random()
        for k in Ruleta:
            if (k[2] >= RandPadre):
                break
        Padre = k

        # Selecciona madre
        RandMadre = random.random()
        for k in Ruleta:
            if (k[2] >= RandMadre):
                break
        Madre = k

        return Padre, Madre, Ruleta, RandPadre, RandMadre

    def Cruce(self, P, M):
        p = P[1]
        m = M[1]

        # Generar punto de cruce al azar
        r = random.randint(3, self.N - 2)

        # Crear hijo 1
        h1 = []
        for k in range(0, r):
            h1.append(p[k])
        for k in range(1, self.N):
            if (m[k] not in h1):
                h1.append(m[k])

        # Crear hijo 2
        h2 = []
        for k in range(0, r):
            h2.append(m[k])
        for k in range(1, self.N):
            if (p[k] not in h2):
                h2.append(p[k])

        return h1, h2,r

    # mutacion swap para el problema de TSP (permutación)
    def Mutacion(self, H, Prob):
        h = H.copy()
        r1 = 0
        r2 = 0
        if (random.random() <= Prob):
            r1 = random.randint(1, self.N - 1)
            r2 = random.randint(1, self.N - 1)
            Aux = h[r1]
            h[r1] = h[r2]
            h[r2] = Aux
        return h, r1, r2

    # Seleccionar N individuos de la elite
    def Elite(self, N):
        # ordenar población por el valor fitnnes
        self.Poblacion = sorted(self.Poblacion, reverse = True)
        PoblacionElite = []
        for k in range(N):
            PoblacionElite.append(self.Poblacion[k])
        return PoblacionElite

    def EjecutarTSP(self, Archivo):
        self.Archivo = Archivo
        # Leer datos del problema
        self.LeerTSP(self.Archivo)

        # Inicializar poblacion inicial con P individuos
        self.Poblacion = self.GeneraPoblacionInicial(self.P)

        # Repetir el proceso evolución por generaciones
        self.HistCosto = []
        self.HistFitness = []
        
        for k in range(self.Generaciones):
            Pe = self.Elite(self.PoElite)
            Po = []

            # obtener padre y madre por la mitad de la población y generar hijo1 e hijo2
            for i in range(self.P // 2):
                Padre, Madre, Ruleta, R1, R2 = self.Seleccion()
                Hijo1, Hijo2, PuntoCruce = self.Cruce(Padre, Madre)
                Hijo1, r1, r2 = self.Mutacion(Hijo1, self.ProbMutacion)
                Hijo2, r1, r2 = self.Mutacion(Hijo2, self.ProbMutacion)
                Po.append([self.Fitnnes(Hijo1), Hijo1])
                Po.append([self.Fitnnes(Hijo2), Hijo2])

            # Reemplazar a la población actual
            self.Poblacion = Po
            # adicionar los individuos de la elite
            for p in Pe:
                self.Poblacion.append(p)
            
            # Guardar costo histórico
            self.Poblacion = sorted(self.Poblacion, reverse=True)
            MejorGene = self.Poblacion[0]
            self.HistCosto.append(self.DistaTotal(MejorGene[1]))
            self.HistFitness.append(self.Fitnnes(MejorGene[1]))
            

        # Elegir el mejor individuo de la población
        self.Poblacion = sorted(self.Poblacion, reverse=True)
        Mejor = self.Poblacion[0]
        self.Ruta = Mejor[1]
        self.Costo = self.DistaTotal(Mejor[1])
        self.Aptitud = Mejor[0]
        return self.Ruta, self.Costo, self.Aptitud 


    def EjecutarDistancia(self, Archivo, Hoja):
        self.Archivo = Archivo
        # Leer datos del problema
        self.LeerDistanciaExcel(self.Archivo, Hoja)

        # Inicializar poblacion inicial con P individuos
        self.Poblacion = self.GeneraPoblacionInicial(self.P)

        # Repetir el proceso evolución por generaciones
        self.HistCosto = []
        self.HistFitness = []
        
        for k in range(self.Generaciones):
            Pe = self.Elite(self.PoElite)
            Po = []

            # obtener padre y madre por la mitad de la población y generar hijo1 e hijo2
            for i in range(self.P // 2):
                Padre, Madre, Ruleta, R1, R2 = self.Seleccion()
                Hijo1, Hijo2, PuntoCruce = self.Cruce(Padre, Madre)
                Hijo1, r1, r2 = self.Mutacion(Hijo1, self.ProbMutacion)
                Hijo2, r1, r2 = self.Mutacion(Hijo2, self.ProbMutacion)
                Po.append([self.Fitnnes(Hijo1), Hijo1])
                Po.append([self.Fitnnes(Hijo2), Hijo2])

            # Reemplazar a la población actual
            self.Poblacion = Po
            # adicionar los individuos de la elite
            for p in Pe:
                self.Poblacion.append(p)
            
            # Guardar costo histórico
            self.Poblacion = sorted(self.Poblacion, reverse=True)
            MejorGene = self.Poblacion[0]
            self.HistCosto.append(self.DistaTotal(MejorGene[1]))
            self.HistFitness.append(self.Fitnnes(MejorGene[1]))
            

        # Elegir el mejor individuo de la población
        self.Poblacion = sorted(self.Poblacion, reverse=True)
        Mejor = self.Poblacion[0]
        self.Ruta = Mejor[1]
        self.Costo = self.DistaTotal(Mejor[1])
        self.Aptitud = Mejor[0]
        return self.Ruta, self.Costo, self.Aptitud 




from meta_lbpt.ag import AlgoritmoGenetico
from meta_lbpt.ch import ColoniaHormiga
from meta_lbpt.at import AlgoritmoTabu
from meta_lbpt.sc import SeleccionClonal
from meta_lbpt.rs import RecocidoSimulado




# -*- coding: utf-8 -*-

"""
# Algoritmo tabú
```
Versión   : 1.2, Para uso educativo
Autor     : Luis Beltran Palma Ttito
Lugar     : Cusco, Perú, 2024.
Proposito : Implementación de algoritmo tabú en python
Problema  : TSP.
```

#Librerías
"""

import random
import math
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

"""# Clase Algoritmo Tabu """

class AlgoritmoTabu:
    def __init__(self, Ciclos):
        self.N = 0                  # Cantidad de ciudades
        self.Ciclos = Ciclos        # Ciclos del algoruitmo tabu
        self.Memoria = []           # Memoria de largo y corto plazo
        self.TSP = []               # TSP
        self.Distancia = []         # matriz de distancias
        self.Archivo = ''           # Archivo TSP

        self.Ruta = []              # Ruta solucion
        self.Costo = 0.0            # Costo de ruta solucion
        self.HistCosto = []         # Historico de costos

    def GraficaEnergia(self):
        plt.figure()
        plt.plot(self.HistCosto)
        plt.title('Evolución de costo')
        plt.grid(True)
        plt.show

    def GraficaRuta(self):
        plt.figure()
        X = []
        Y = []
        for i in range(len(self.Ruta)):
            X.append(self.TSP[self.Ruta[i]-1, 0])
            Y.append(self.TSP[self.Ruta[i]-1, 1])
        plt.plot(X,Y,'-o')
        for i in range(len(self.Ruta)):
            plt.annotate(str(self.Ruta[i]), (X[i],Y[i]))
        plt.title("Ruta solución")
        plt.show()

    # Lee Matriz desde excel (sin cabecera) y crea matriz de distancias
    def LeerDistanciaExcel(self, Archivo, Hoja ):
        # Leer TSP
        df = pd.read_excel(Archivo, sheet_name=Hoja, header=None)
        df_float = df.astype(float)
        self.Distancia = np.array(df_float.values)
        self.N = len(self.Distancia)
        self.Memoria = np.zeros((self.N, self.N), dtype=int)
        return self.N, self.TSP, self.Distancia
    
    # Lee archivo TSP descargado de TSPLIB
    def Leer_TSP(self, Archivo):
        with open(Archivo) as archivo:
            ciudades = []
            while True:
                linea = archivo.readline()
                if linea.startswith("NODE_COORD_SECTION"):
                    break
            for linea in archivo:
                partes = linea.strip().split()
                if partes[0] == "EOF":
                    break
                _, x, y = partes
                ciudades.append([float(x), float(y)])
        return np.array(ciudades)

    # Lee Archivo TSP y crea matriz de distancias
    def LeerTSP(self, Archivo):
        # Leer TSP
        self.TSP = self.Leer_TSP(Archivo)
        self.N = len(self.TSP)
        # Convertir TSP a matriz de distancias
        self.Distancia = np.empty((self.N,self.N))
        for i in range(self.N):
            for j in range(self.N):
                if i == j:
                    self.Distancia[i, j] = 0
                else:
                    self.Distancia[i, j] = math.sqrt(math.pow(self.TSP[i, 0] - self.TSP[j, 0], 2)
                    + math.pow(self.TSP[i, 1] - self.TSP[j, 1], 2))
        
        self.Memoria = np.zeros((self.N, self.N), dtype=int)
        return self.N, self.TSP, self.Distancia


    # Determina la distancia entre 2 ciudad, usa la matriz Distancias
    def CostoArco(self, inicio, fin):
        return self.Distancia[inicio - 1][fin - 1]

    # Genera un estado inicial
    def Muestrar(self):
        So = random.sample(list(range(2,self.N+1)), k=self.N-1)
        So.insert(0, 1)
        return So

    # Determina la energía de S (heurístico), costo de la ruta S
    def Energia(self, S):
        Costo = 0
        for k in range(self.N - 1):
            Costo = Costo + self.CostoArco(S[k], S[k+1])
        Costo = Costo + self.CostoArco(S[self.N - 1], 1)
        return Costo

    # Determina los vecinos de una estado
    # S: estado del que se desea conocer los vecinos
    # Lista, cada elemento contiene:
    # 1. puntos que favorece al vecino en comparación al estado S
    # 2. estado 1 permutado
    # 3. estado 2 permutado
    # 4. Estado vecino
    def Candidatos(self, S):
        Lista = []
        for k in range(1, self.N):
            for j in range(k + 1, self.N):
                L = []
                SS = S.copy()
                Aux = SS[k]
                SS[k] = SS[j]
                SS[j] = Aux
                L.append(self.Energia(S) - self.Energia(SS))
                L.append(k+1)
                L.append(j+1)
                L.append(SS)
                Lista.append(L)
        Lista = sorted(Lista, reverse=True)
        return Lista

    def Tabu(self, S):
        Candi = self.Candidatos(S)
        k = 0

        # Verificar que estado no es tabú
        d = Candi[k]
        while (self.Memoria[d[1]-1][d[2]-1] > 0 or  k < self.N):
            k = k + 1
            d = Candi[k]

        # Elige el ultimo analizado o el primero
        if (k < self.N):
            d = Candi[k]
        else:
            d = Candi[0]

        # Actualizar situación de tabu (-1)
        for i in range(0, self.N - 1):
            for j in range(i + 1, self.N - 1):
                if (self.Memoria[i][j] > 0):
                    self.Memoria[i][j] = self.Memoria[i][j] - 1

        # Actualiza memoria a largo plazo
        if (self.Memoria[d[1]-1][d[2]-1] == 0):
            self.Memoria[d[1]-1][d[2]-1] = 3   # memoria a corto plazo
            self.Memoria[d[2]-1][d[1]-1] = self.Memoria[d[2]-1][d[1]-1]  + 1  # memoria a largo plazo
            S = (d[3]).copy()

        return S, Candi, self.Memoria

    def EjecutarTSP(self, Archivo):
        self.Archivo = Archivo
        self.LeerTSP(self.Archivo)
        S = self.Muestrar()
        self.HistCosto = []
        for k in range(self.Ciclos):
            S, Candi, self.Memoria = self.Tabu(S)
            self.HistCosto.append(self.Energia(S))
        self.Ruta = S
        self.Costo = self.Energia(S)
        return self.Ruta, self.Costo

    def EjecutarDistancia(self, Archivo, Hoja):
        self.Archivo = Archivo
        self.LeerDistanciaExcel(self.Archivo, Hoja)
        S = self.Muestrar()
        self.HistCosto = []
        for k in range(self.Ciclos):
            S, Candi, self.Memoria = self.Tabu(S)
            self.HistCosto.append(self.Energia(S))
        self.Ruta = S
        self.Costo = self.Energia(S)
        return self.Ruta, self.Costo


# -*- coding: utf-8 -*-

"""
# Selección clonal
```
Versión   : 1.2, Para uso educativo
Autor     : Luis Beltran Palma Ttito
Lugar     : Cusco, Perú, 2024.
Proposito : Implementación de selección clonal en python
Problema  : TSP.
```

#Librerías
"""

import random
import math
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

"""# Clase Selección Clonal """

class SeleccionClonal:
    def __init__(self, Iteraciones, Poblacion, CantClones, AltaAfinidad, ProbMutacion, IncProbMutacion):
        self.Iteraciones = Iteraciones          # Iteraciones
        self.P = Poblacion                      # Población de anticuerpos
        self.CantClones = CantClones            # Cantidad de clones
        self.Q = AltaAfinidad                   # Cantidad de anticuerpos de alta afinidad
        self.ProbMutar = ProbMutacion           # Probabiliad de mutar para individuos más afines
        self.IncMutacion = IncProbMutacion      # Incremento del 150% de mutacion a perores individuos
        self.N = 0                              # Cantidad de ciudades
        self.Distancia = []                     # Matriz de distancias 
        self.Archivo = ''                       # Archivo TSP
        self.TSP = []                           # TSP

        self.MejorObjetivo = 0          # Mejor objetivo hallado
        self.Ruta = []                  # Ruta solucion
        self.Costo = 0.0                # Costo de ruta solucion 
        self.HistCosto = []             # Historico de costos

    # Muestra evolución de costo
    def GraficaAfinidad(self):
        plt.figure()
        plt.plot(self.HistCosto)
        plt.title('Evolución de costo')
        plt.grid(True)
        plt.show

    # Grafica la ruta solución 2D
    def GraficaRuta(self):
        plt.figure()
        X = []
        Y = []
        for i in range(len(self.Ruta)):
            X.append(self.TSP[self.Ruta[i]-1, 0])
            Y.append(self.TSP[self.Ruta[i]-1, 1])
        plt.plot(X,Y,'-o')
        for i in range(len(self.Ruta)):
            plt.annotate(str(self.Ruta[i]), (X[i],Y[i]))
        plt.title("Ruta solución")
        plt.show()

    # Lee Matriz desde excel (sin cabecera) y crea matriz de distancias
    def LeerDistanciaExcel(self, Archivo, Hoja ):
        # Leer TSP
        df = pd.read_excel(Archivo, sheet_name=Hoja, header=None)
        df_float = df.astype(float)
        self.Distancia = np.array(df_float.values)
        self.N = len(self.Distancia)
        return self.N, self.TSP, self.Distancia
    
    # Lee archivo TSP descargado de TSPLIB
    def Leer_TSP(self, Archivo):
        with open(Archivo) as archivo:
            ciudades = []
            while True:
                linea = archivo.readline()
                if linea.startswith("NODE_COORD_SECTION"):
                    break
            for linea in archivo:
                partes = linea.strip().split()
                if partes[0] == "EOF":
                    break
                _, x, y = partes
                ciudades.append([float(x), float(y)])
        return np.array(ciudades)

    # Lee Archivo TSP y crea matriz de distancias
    def LeerTSP(self, Archivo):
        # Leer TSP
        self.TSP = self.Leer_TSP(Archivo)
        self.N = len(self.TSP)
        # Convertir TSP a matriz de distancias
        self.Distancia = np.empty((self.N,self.N))
        for i in range(self.N):
            for j in range(self.N):
                if i == j:
                    self.Distancia[i, j] = 0
                else:
                    self.Distancia[i, j] = math.sqrt(math.pow(self.TSP[i, 0] - self.TSP[j, 0], 2)
                    + math.pow(self.TSP[i, 1] - self.TSP[j, 1], 2))
        return self.N, self.TSP, self.Distancia

    # Determina la distancia entre 2 ciudad, usa la matriz D
    def CostoArco(self, inicio, fin):
        return self.Distancia[inicio - 1][fin - 1]

    # Determina la energía de S (heurístico), costo de la ruta S
    def Objetivo(self, S):
        Costo = 0
        for k in range(self.N - 1):
            Costo = Costo + self.CostoArco(S[k],S[k+1])
        Costo = Costo + self.CostoArco(S[self.N - 1], 1)
        return Costo

    # Genera población inicial con P individuos
    def GeneraPoblacionInicial(self, P):
        self.Poblacion = []
        self.P = P
        for k in range(1, P + 1):
            S = random.sample(list(range(2,self.N+1)), self.N-1)
            S.insert(0, 1)
            self.Poblacion.append([self.Objetivo(S),S])
        self.Poblacion = sorted(self.Poblacion)
        Mejor = self.Poblacion[0]
        self.MejorObjetivo = Mejor[0]
        return self.Poblacion

    # Obtener los Q individuos de mas alta afinidad
    def AltaAfinidad(self, Q):
        self.PoblacionAltaAfinidad = []
        self.Poblacion = sorted(self.Poblacion)
        for k in range(Q):
            C = (self.Poblacion[k]).copy()
            self.PoblacionAltaAfinidad.append(C)
        return self.PoblacionAltaAfinidad

    # Mutar con probabiliad Prob el individuo I
    def Mutar(self, I, Prob):
        IM = I.copy()
        if (random.random() <= Prob):
            r1 = random.randint(1, self.N - 1)
            r2 = random.randint(1, self.N - 1)
            Aux = IM[r1]
            IM[r1] = IM[r2]
            IM[r2] = Aux
        return IM

    # Clonar y mutar M muestras de cada individuo de alta afinidad
    # fp: probabilidad de mutación
    # fi: incremento de prob de mutacion
    def ClonarMutar(self, M, fp, fi):
        self.ProbMutar = fp
        self.IncMutacion = fi
        self.C = []
        self.PoblacionAltaAfinidad = sorted(self.PoblacionAltaAfinidad)
        Prob = self.ProbMutar
        for IAA in self.PoblacionAltaAfinidad:
            for k in range(M):
                IA = self.Mutar(IAA[1], Prob)
                self.C.append([self.Objetivo(IA), IA])
                Prob = Prob * self.IncMutacion  # incrementa probabiliad de mutación
        return self.C

    # Elegir soluciones que optimizan soluciones anteriores
    def ElegirOptimizadores(self):
        self.Optimos = []
        self.C = sorted(self.C)
        for I in self.C:
            if (I[0] <= self.MejorObjetivo):
                self.Optimos.append(I)
            else:
                break
        return self.Optimos

    def ReemplazarOptimos(self):
        # Reemplazar los peores por  mejores
        self.Poblacion = sorted(self.Poblacion, reverse = True)

        k = 0
        for IO in self.Optimos:
            if k >= len(self.Poblacion):
                break
            self.Poblacion[k] = IO
            k = k + 1

        # Retornar la nueva población
        return self.Poblacion

    def EjecutarTSP(self, Archivo):
        self.Archivo = Archivo
        self.LeerTSP(self.Archivo)
        self.GeneraPoblacionInicial(self.P)
        self.HistCosto = []
        for k in range(self.Iteraciones):
            self.AltaAfinidad(self.Q)
            self.ClonarMutar(self.CantClones, self.ProbMutar, self.IncMutacion)
            self.ElegirOptimizadores()
            self.ReemplazarOptimos()
            # Calculo de historico de afinidad 
            self.Poblacion = sorted(self.Poblacion)
            Mejor = self.Poblacion[0]
            self.HistCosto.append(Mejor[0])

        self.Poblacion = sorted(self.Poblacion)
        Solucion = self.Poblacion[0]
        self.Ruta = Solucion[1]
        self.Costo =  Solucion[0]
        return self.Ruta, self.Costo


    def EjecutarDistancia(self, Archivo, Hoja):
        self.Archivo = Archivo
        self.LeerDistanciaExcel(self.Archivo, Hoja)
        self.GeneraPoblacionInicial(self.P)
        self.HistCosto = []
        for k in range(self.Iteraciones):
            self.AltaAfinidad(self.Q)
            self.ClonarMutar(self.CantClones, self.ProbMutar, self.IncMutacion)
            self.ElegirOptimizadores()
            self.ReemplazarOptimos()
            # Calculo de historico de afinidad 
            self.Poblacion = sorted(self.Poblacion)
            Mejor = self.Poblacion[0]
            self.HistCosto.append(Mejor[0])

        self.Poblacion = sorted(self.Poblacion)
        Solucion = self.Poblacion[0]
        self.Ruta = Solucion[1]
        self.Costo =  Solucion[0]
        return self.Ruta, self.Costo






import pathlib
from setuptools import find_packages, setup

HERE = pathlib.Path(__file__).parent

LONG_DESCRIPTION = """

# **Algoritmos Metaheurísticos para el problema TSP**


*luis.palma@unsaac.edu.pe, Luis Beltran Palma Ttito, Universidad Nacional de San Antonio Abad del Cusco (Autor)*

*dennis.candia@unsaac.edu.pe, Dennis Iván Candia Oviedo, Universidad Nacional de San Antonio Abad del Cusco (Coautor)*

*carlos.quispe@unsaac.edu.pe, Carlos Ramón Quispe Onofre, Universidad Nacional de San Antonio Abad del Cusco (Coautor)*

*willian.zamalloa@unsaac.edu.pe, Willian Zamalloa Paro, Universidad Nacional de San Antonio Abad del Cusco (Coautor)*

---

La presente librería implementa los algoritmos metaheurísticos siguientes:
1. Algoritmo genético
2. Colonia de hormigas
3. Algoritmo tabú
4. Selección clonal
5. Recocido simulado


#### Propósito del la librería
> La presente liberería, fue creada con motivo educativos, y únicamente resuelve
el problema de TSP, tomando en cuenta las siguientes consideraciones:
1. Los algoritmos metaheurísticos utilizan un vector con N elementos, para almacenar la ruta a seguir en las N ciudades.
2. Los valores del vector nunca se repiten, el primer valor siempre es 1 (se asume que el vendedor parte de la ciudad 1)
3. Los datos de entrada deben ser del tipo TSP (coordenadas de ciudades en 2D)
4. Los archivos TSP pueden ser descargados de TSPLIB http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/tsp/
5. Tambien es posible utilizar de entrada Matrices de Distancias en archivo  Excel, sin cabecera y sin indice (en caso de usar matriz de distancias, no es posible graficar la ruta) 

*luis.palma@unsaac.edu.pe, Luis Beltran Palma Ttito, Universidad Nacional de San Antonio Abad del Cusco*

---

## Problema TSP "Travelling Salesman Problem" (Problema del Vendedor Viajante)
*Un viajero de comercio parte de una ciudad y las distancias a otras ciudades son conocidas, ¿cuál es la **ruta óptima** que debe seguir para visitar todas las ciudades y volver a la ciudad de partida?*

---

## **0. Pasos previos**

**Paso 1**: Instalación del la librería de algoritmos metaheuristicos
```
pip install meta-lbpt
```
**Paso 2**: Importación de librerías
```
from meta_lbpt.ag import AlgoritmoGenetico
from meta_lbpt.ch import ColoniaHormiga
from meta_lbpt.at import AlgoritmoTabu
from meta_lbpt.sc import SeleccionClonal
from meta_lbpt.rs import RecocidoSimulado
```

---

## **1. Algoritmo genético**

**Paso 1.1.** Ejecutar el algoritmo genético con TSP
```
# Parametros: Generaciones, Población, ProbMutación, PoblacionElite
# burma14.tsp: Archivo que contiene las N ciudades en formato TSP (coordenadas 2D)
AG = AlgoritmoGenetico(100, 500, 0.05, 2)
Solucion, Costo, Aptitud = AG.EjecutarTSP('burma14.tsp')
print('Ruta encontrada por el algoritmo genético: ', Solucion)
print('Costo: ', Costo)
print('Aptitud: ', Aptitud)
```

```
Ruta encontrada por el algoritmo genético:  [1, 2, 10, 9, 11, 8, 13, 7, 12, 6, 5, 4, 3, 14]
Costo:  32.44685100783281
Aptitud:  0.03081963176514712
```

**Paso 1.2.** Graficar la ruta solución
```
AG.GraficaRuta()
```


**Paso 1.3** Gráfica la evolución de costo
```
AG.GraficaCosto()
```

**Paso 1.4.** Ejecutar el algoritmo genético con Matriz de Distancias
```
# Parametros: Generaciones, Población, ProbMutación, PoblacionElite
# burma14.tsp: Archivo que contiene las N ciudades en formato TSP (coordenadas 2D)
AG = AlgoritmoGenetico(100, 500, 0.05, 2)
Solucion, Costo, Aptitud = AG.EjecutarDistancia('distancia.xlsx', 'Hoja1')
print('Ruta encontrada por el algoritmo genético: ', Solucion)
print('Costo: ', Costo)
print('Aptitud: ', Aptitud)
```

---

## **2. Colonia de hormigas**

**Paso 2.1.** Ejecutar el algoritmo colonia de hormigas con TSP
```
# Parametros: Iteraciones, PoblacionHormigas, Coef. evaporación, Coef. aprendizaje
# burma14.tsp: Archivo que contiene las ciudades en formato TSP
CH = ColoniaHormiga(50, 150, 0.061, 0.65)
Solucion, Costo = CH.EjecutarTSP('burma14.tsp')
print('Mejor ruta encontrada por la colonia de hormigas: ', Solucion)
print('Costo: ',Costo)
```
```
Mejor ruta encontrada por la colonia de hormigas:  [1, 2, 3, 4, 5, 6, 12, 14, 7, 13, 9, 11, 10, 8]
Costo:  32.76917155355312
```

**Paso 2.2.** Gráfica la ruta solución
```
CH.GraficaRuta()
```

**Paso 2.3.** Gráfica la evolución de costo
```
CH.GraficaCosto()
```
**Paso 2.4.** Ejecutar el algoritmo colonia de hormigas con Matriz de Distancias
```
# Parametros: Iteraciones, PoblacionHormigas, Coef. evaporación, Coef. aprendizaje
# burma14.tsp: Archivo que contiene las ciudades en formato TSP
CH = ColoniaHormiga(50, 150, 0.061, 0.65)
Solucion, Costo = CH.EjecutarDistancia('distancia.xlsx', 'Hoja1')
print('Mejor ruta encontrada por la colonia de hormigas: ', Solucion)
print('Costo: ',Costo)
```
---

## **3. Algoritmo tabú**

**Paso 3.1.** Ejecutar el algoritmo tabú con TSP
```
# Parametros: Ciclos
# burma14.tsp: Archivo que contiene ciudades en formato TSP
AT = AlgoritmoTabu(20)
Solucion, Energia= AT.EjecutarTSP('burma14.tsp')
print('Mejor ruta encontrada por algorítmo tabú: ', Solucion)
print('Energía: ', Energia)
```
```
Mejor ruta encontrada por algorítmo tabú:  [1, 8, 13, 7, 12, 6, 5, 4, 3, 14, 2, 10, 9, 11]
Energía:  31.226915109427544
```
**Paso 3.2.** Graficar la ruta solución
```
AT.GraficaRuta()
```

**Paso 3.3.** Graficar la evolución de costo
```
AT.GraficaEnergia()
```

**Paso 3.4.** Ejecutar el algoritmo tabú con TSP con Matriz de Distancias
```
# Parametros: Ciclos
# burma14.tsp: Archivo que contiene ciudades en formato TSP
AT = AlgoritmoTabu(20)
Solucion, Energia= AT.EjecutarDistancia('distancia.xlsx', 'Hoja1')
print('Mejor ruta encontrada por algorítmo tabú: ', Solucion)
print('Energía: ', Energia)
```

---

## **4. Selección clonal**

**Paso 4.1.** Ejecutar el algoritmo selección clonal con TSP
```
# Parametros: Iteraciones, Población, CantClones, CantIndividuoAltaAfinidad, ProbMutacion, IncProbMutacion
# burma14.tsp: Archivo con datos de ciudades en formato TSP
SC = SeleccionClonal(30, 150, 10, 5, 0.8, 1.38)
Solucion, Afinidad = SC.EjecutarTSP('burma14.tsp')
print('Ruta solución: ', Solucion)
print('Afinidad: ', Afinidad)
```
```
Ruta solución:  [1, 9, 10, 2, 14, 3, 4, 5, 6, 12, 7, 13, 11, 8]
Afinidad:  32.81277804469343
```
**Paso 4.2.** Graficar la ruta solución
```
SC.GraficaRuta()
```

**Paso 4.3.** Graficar la evolución de costo
```
SC.GraficaAfinidad()
```

**Paso 4.4.** Ejecutar el algoritmo selección clonal con Matriz de Distancias
```
# Parametros: Iteraciones, Población, CantClones, CantIndividuoAltaAfinidad, ProbMutacion, IncProbMutacion
# burma14.tsp: Archivo con datos de ciudades en formato TSP
SC = SeleccionClonal(30, 150, 10, 5, 0.8, 1.38)
Solucion, Afinidad = SC.EjecutarDistancia('distancia.xlsx', 'Hoja1')
print('Ruta solución: ', Solucion)
print('Afinidad: ', Afinidad)
```

---

## **5. Recocido simulado**

**Paso 5.1.** Ejecutar el algoritmo recocido simulado con TSP
```
# Parametros: TempInicial, TempFinal, ConstanteDecremento, TiempoEnfriamiento, Cte. Bolztman
# burma14.tsp: Archivo con ciudades en formato TSP
RS = RecocidoSimulado(0.99, 0.02, 0.99, 44, 0.97)
Solucion, Costo = RS.EjecutarTSP('burma14.tsp')
print('Mejor solución obtenido por enfriamiento simulado: ',Solucion)
print('Costo: ',Costo)
```
```
Mejor solución obtenido por enfriamiento simulado:  [1, 10, 9, 11, 8, 13, 7, 14, 12, 6, 5, 4, 3, 2]
Costo:  31.828317520454547
```

**Paso 5.2.** Graficar la ruta solución
```
RS.GraficaRuta()
```

**Paso 5.3.** Graficar la evolución de costo
```
RS.GraficaEnergia()
```

**Paso 5.4.** Ejecutar el algoritmo recocido simulado con Matriz de Distancias
```
# Parametros: TempInicial, TempFinal, ConstanteDecremento, TiempoEnfriamiento, Cte. Bolztman
# burma14.tsp: Archivo con ciudades en formato TSP
RS = RecocidoSimulado(0.99, 0.02, 0.99, 44, 0.97)
Solucion, Costo = RS.EjecutarDistancia('distancia.xlsx', 'Hoja1')
print('Mejor solución obtenido por enfriamiento simulado: ',Solucion)
print('Costo: ',Costo)
```

---

En caso de existir consultas, escribanos al email: 

luis.palma@unsaac.edu.pe, Luis Beltran Palma Ttito (autor)

dennis.candia@unsaac.edu.pe, Dennis Iván Candia Oviedo (Coautor)*

carlos.quispe@unsaac.edu.pe, Carlos Ramón Quispe Onofre (Coautor)*

willian.zamalloa@unsaac.edu.pe, Willian Zamalloa Paro (Coautor)*


"""

setup(
    name='meta_lbpt',
    version='1.0.7',
    description='Algoritmos Metaheurísticos',
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Luis Beltran Palma Ttito, Dennis Iván Candia Oviedo, "
           "Carlos Ramón Quispe Onofre, Willian Zamalloa Paro",
    author_email="luis.palma@unsaac.edu.pe, dennis.candia@unsaac.edu.pe, "
                 "carlos.quispe@unsaac.edu.pe, willian.zamalloa@unsaac.edu.pe",
    license='GPL',
    packages=find_packages(),
    include_package_data=True
)


# -*- coding: utf-8 -*-

"""
# Recocido simulado
```
Versión   : 1.2, Para uso educativo
Autor     : Luis Beltran Palma Ttito
Lugar     : Cusco, Perú, 2024.
Proposito : Implementación de recocido simulado en python
Problema  : TSP.
```

#Librerías
"""

import random
import math
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np

"""# Clase Recocido Simulado """


class RecocidoSimulado:
    def __init__(self, TempInicial, TempFinal, ConstanteDecremento, TiempoEnfriamiento, Bolztman):
        self.Ti = TempInicial           # Temperatura inicial
        self.Tf = TempFinal             # Temperatura final
        self.c = ConstanteDecremento    # Constante de decremento
        self.n = TiempoEnfriamiento     # Duración de tiempo de enfriamiento
        self.k = Bolztman               # Constante de Bolztman
        self.TSP = []                   # TSP
        self.Distancia = []             # matriz de distancias
        self.N = 0                      # Cantidad de ciudades
        self.Archivo = ''               # Archivo TSP

        self.Ruta = []          # Ruta solucion
        self.Costo = 0.0        # costo de la ruta solucion
        self.HistCosto = []     # Historico de costo

    # Muestra evolución de costo
    def GraficaEnergia(self):
        plt.figure()
        plt.plot(self.HistCosto)
        plt.title('Evolución de costo')
        plt.grid(True)
        plt.show

    # Grafica la ruta solución 2D
    def GraficaRuta(self):
        plt.figure()
        X = []
        Y = []
        for i in range(len(self.Ruta)):
            X.append(self.TSP[self.Ruta[i]-1, 0])
            Y.append(self.TSP[self.Ruta[i]-1, 1])
        plt.plot(X,Y,'-o')
        for i in range(len(self.Ruta)):
            plt.annotate(str(self.Ruta[i]), (X[i],Y[i]))
        plt.title("Ruta solución")
        plt.show()

    # Lee Matriz desde excel (sin cabecera) y crea matriz de distancias
    def LeerDistanciaExcel(self, Archivo, Hoja ):
        # Leer TSP
        df = pd.read_excel(Archivo, sheet_name=Hoja, header=None)
        df_float = df.astype(float)
        self.Distancia = np.array(df_float.values)
        self.N = len(self.Distancia)
        return self.N, self.TSP, self.Distancia
    
    # Lee archivo TSP descargado de TSPLIB
    def Leer_TSP(self, Archivo):
        with open(Archivo) as archivo:
            ciudades = []
            while True:
                linea = archivo.readline()
                if linea.startswith("NODE_COORD_SECTION"):
                    break
            for linea in archivo:
                partes = linea.strip().split()
                if partes[0] == "EOF":
                    break
                _, x, y = partes
                ciudades.append([float(x), float(y)])
        return np.array(ciudades)
      
    # Lee Archivo TSP y crea matriz de distancias
    def LeerTSP(self, Archivo):
        # Leer TSP
        self.TSP = self.Leer_TSP(Archivo)
        self.N = len(self.TSP)
        # Convertir TSP a matriz de distancias
        self.Distancia = np.empty((self.N,self.N))
        for i in range(self.N):
            for j in range(self.N):
                if i == j:
                    self.Distancia[i, j] = 0
                else:
                    self.Distancia[i, j] = math.sqrt(math.pow(self.TSP[i, 0] - self.TSP[j, 0], 2)
                    + math.pow(self.TSP[i, 1] - self.TSP[j, 1], 2))

        return self.N, self.TSP, self.Distancia

    # Determina la distancia entre 2 ciudad, usa la matriz D
    def CostoArco(self, inicio, fin):
        return self.Distancia[inicio - 1][fin - 1]

    # Determina la energía de S (heurístico), costo de la ruta S
    # [1, 6, 8, 2, 10, 3, ..]
    def Energia(self, S):
        Cost = 0
        for k in range(self.N - 1):
            Cost = Cost + self.CostoArco(S[k],S[k+1])
        Cost = Cost + self.CostoArco(S[self.N - 1], 1)
        return Cost

    # Determina un solución aleatoria inicial
    def Muestrar(self):
        So = random.sample(list(range(2,self.N+1)), self.N-1)
        So.insert(0, 1)
        return So

    # Determina un vecino al azar
    def Mover(self, S):
        SS = S.copy()
        Pos1 = int(random.uniform(1, self.N))
        Pos2 = int(random.uniform(1, self.N))
        Aux = SS[Pos1]
        SS[Pos1] = SS[Pos2]
        SS[Pos2] = Aux
        return SS

    # Algoritmo de recocido simulado
    def EjecutarTSP(self, Archivo):
        self.Archivo = Archivo
        self.LeerTSP(self.Archivo)
        # variables que almacenan rutas y energías
        Rutas = []
        Energias = []
        # Inicializa T con temperatura inicial
        T = self.Ti
        # Generar la primera solución
        x = self.Muestrar()
        # calcula energía de la muestra inicial
        e = self.Energia(x)
        # alamacena en la lista
        Energias.append(e)
        Rutas.append(x)

        # Mientras la temperatura actual > a temperatura final
        while (T > self.Tf):
            # por n tiempos determina un mejor solución
            # n tiempos de emfriamiento
            for i in range(1, self.n + 1):
                # Genera vecino
                xp = self.Mover(x)
                ep = self.Energia(xp)
                # si energia de vecino < energia de dato actual, entonces reemplazar
                if (ep < e):
                    x = xp
                    e = ep
                    Energias.append(e)
                    Rutas.append(x)
                # Utilizar clave de funcionamiento de enfriamiento simulado
                # para escapar de minimos locales
                # self.k: constante de Boltzmann
                else:
                    r = random.random()
                    if (r < math.exp(-(ep-e)/(self.k * T))):
                        x = xp
                        e = ep
                        Energias.append(e)
                        Rutas.append(x)

            # Disminuir temperatura
            T = self.c * T
            self.Costo = self.Energia(x)
            self.Ruta = x
            self.HistCosto = Energias
        return self.Ruta, self.Costo
    
    # Algoritmo de recocido simulado
    def EjecutarDistancia(self, Archivo, Hoja):
        self.Archivo = Archivo
        self.LeerDistanciaExcel(self.Archivo, Hoja)
        # variables que almacenan rutas y energías
        Rutas = []
        Energias = []
        # Inicializa T con temperatura inicial
        T = self.Ti
        # Generar la primera solución
        x = self.Muestrar()
        # calcula energía de la muestra inicial
        e = self.Energia(x)
        # alamacena en la lista
        Energias.append(e)
        Rutas.append(x)

        # Mientras la temperatura actual > a temperatura final
        while (T > self.Tf):
            # por n tiempos determina un mejor solución
            # n tiempos de emfriamiento
            for i in range(1, self.n + 1):
                # Genera vecino
                xp = self.Mover(x)
                ep = self.Energia(xp)
                # si energia de vecino < energia de dato actual, entonces reemplazar
                if (ep < e):
                    x = xp
                    e = ep
                    Energias.append(e)
                    Rutas.append(x)
                # Utilizar clave de funcionamiento de enfriamiento simulado
                # para escapar de minimos locales
                # self.k: constante de Boltzmann
                else:
                    r = random.random()
                    if (r < math.exp(-(ep-e)/(self.k * T))):
                        x = xp
                        e = ep
                        Energias.append(e)
                        Rutas.append(x)

            # Disminuir temperatura
            T = self.c * T
            self.Costo = self.Energia(x)
            self.Ruta = x
            self.HistCosto = Energias
        return self.Ruta, self.Costo
    